"""This module contains a decorator :func:`cached` that can be used to
cache the results of any Python functions to disk.

This is useful when you have functions that take a long time to
compute their value, and you want to cache the results of those
functions between runs.

Python's :mod:`pickle` is used to serialize data.  All cache files go
into the `cache/` directory inside your working directory.

`@cached` uses a cache key function to find out if it has the value
for some given function arguments cached on disk.  The way it
calculates that cache key by default is to simply use the string
representation of all arguments passed into the function.  Thus, the
default cache key function looks like this:

.. code-block:: python

    def default_cache_key(*args, **kwargs):
        return str(args) + str(sorted(kwargs.items()))

Here is an example use of the :func:`cached` decorator:

.. code-block:: python

    import math
    @cached()
    def fac(x):
        print 'called!'
        return math.factorial(x)

    fac(20)
    called!
    2432902008176640000
    fac(20)
    2432902008176640000

Often you will want to use a more intelligent cache key, one that
takes more things into account.  Here's an example cache key function
for a cache decorator used with a `transform` method of a scikit-learn
:class:`~sklearn.base.BaseEstimator`:

.. doctest::

    >>> def transform_cache_key(self, X):
    ...     return ','.join([
    ...         str(X[:20]),
    ...         str(X[-20:]),
    ...         str(X.shape),
    ...         str(sorted(self.get_params().items())),
    ...         ])

This function puts the first and the last twenty rows of the matrix
`X` into the cache key.  On top of that, it adds the shape of the
matrix `X.shape` along with the items in `self.get_params`, which with
a scikit-learn :class:`~sklearn.base.BaseEstimator` class is the
dictionary of model parameters.  This makes sure that even though the
input matrix is the same, it will still calculate the value again if
the value of `self.get_params()` is different.

Your estimator class can then use the decorator like so:

.. code-block:: python

    class MyEstimator(BaseEstimator):
        @cached(transform_cache_key)
        def transform(self, X):
            # ...
"""

import cPickle
from functools import wraps
import hashlib
import logging
import random
import os
import string


CACHE_PATH = 'cache/'
if not os.path.exists(CACHE_PATH):  # pragma: no cover
    os.mkdir(CACHE_PATH)

logger = logging.getLogger(__name__)


def default_cache_key(*args, **kwargs):
    return str(args) + str(sorted(kwargs.items()))


def cached(cache_key=default_cache_key, cache_path=None):
    def cached(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            # Calculation of the cache key is delegated to a function
            # that's passed in via the decorator call
            # (`default_cache_key` by default).
            key = str(cache_key(*args, **kwargs))
            hashed_key = hashlib.sha1(key).hexdigest()[:8]

            # We construct the filename using the cache key.  If the
            # file exists, unpickle and return the value.
            filename = os.path.join(
                cache_path or CACHE_PATH,
                '{}.{}-cache-{}.pkl'.format(
                    func.__module__, func.__name__, hashed_key))

            if os.path.exists(filename):
                filesize = os.path.getsize(filename)
                size = "%0.1f MB" % (filesize / (1024 * 1024.0))
                logger.debug(" * cache hit: {} ({})".format(filename, size))
                with open(filename, 'r') as f:
                    return cPickle.load(f)
            else:
                logger.debug(" * cache miss: {}".format(filename))
                value = func(*args, **kwargs)
                tmp_filename = '{}-{}.tmp'.format(
                    filename,
                    ''.join(random.sample(string.ascii_letters, 4)),
                    )
                with open(tmp_filename, 'wb') as f:
                    try:
                        cPickle.dump(value, f, -1)
                        os.rename(tmp_filename, filename)
                    except SystemError:
                        logger.warning(
                            "Saving pickle {} resulted in SystemError".format(
                            filename))
                return value

        wrapper.uncached = func
        return wrapper
    return cached



from mock import patch


def test_cached(tmpdir):
    from ..cache import cached

    called = []

    @cached(cache_path=str(tmpdir))
    def add(one, two):
        called.append([one, two])
        return one + two

    assert add(2, 3) == 5
    assert len(called) == 1
    assert add(2, 3) == 5
    assert len(called) == 1
    assert add(2, 4) == 6
    assert len(called) == 2


def test_cache_with_cache_key(tmpdir):
    from ..cache import cached

    called = []

    def cache_key(one, two):
        return one

    @cached(cache_key=cache_key, cache_path=str(tmpdir))
    def add(one, two):
        called.append([one, two])
        return one + two

    assert add(2, 3) == 5
    assert len(called) == 1
    assert add(2, 3) == 5
    assert len(called) == 1
    assert add(2, 4) == 5
    assert len(called) == 1


def test_cache_systemerror(tmpdir):
    from ..cache import cached

    @cached(cache_path=str(tmpdir))
    def add(one, two):
        return one + two

    with patch('nolearn.cache.cPickle.dump') as dump:
        dump.side_effect = SystemError()
        assert add(2, 3) == 5



from mock import Mock
import numpy as np
from scipy import sparse


def test_abstract_model():
    from ..model import AbstractModel

    pipeline = Mock()

    class MyModel(AbstractModel):
        default_params = {'param1': 1}
        @property
        def pipeline(self):
            return pipeline

    model = MyModel(param2=2)
    assert model() == pipeline

    pipeline.set_params.assert_called_with(param1=1, param2=2)


def test_averaging_estimator_fit():
    from ..model import AveragingEstimator

    estimator1 = Mock()
    estimator2 = Mock()
    X, y = Mock(), Mock()

    averaging = AveragingEstimator([estimator1, estimator2], verbose=1)
    assert averaging.fit(X, y) == averaging

    assert averaging.estimators == [
        estimator1.fit.return_value,
        estimator2.fit.return_value,
        ]

    estimator1.fit.assert_called_with(X, y)
    estimator2.fit.assert_called_with(X, y)


def test_averaging_estimator_predict():
    from ..model import AveragingEstimator

    estimator1 = Mock()
    estimator2 = Mock()
    X = Mock()

    estimator1.predict_proba.return_value = np.array([[0.1, 0.2, 0.3]])
    estimator2.predict_proba.return_value = np.array([[0.4, 0.5, 0.6]])

    averaging = AveragingEstimator([estimator1, estimator2], verbose=1)
    assert averaging.predict(X) == [2]

    estimator1.predict_proba.assert_called_with(X)
    estimator2.predict_proba.assert_called_with(X)

    assert (averaging.predict_proba(X) == [[0.45, 0.6, 0.75]]).all()


class TestFeatureStacker(object):
    def test_fit(self):
        from ..model import FeatureStacker

        first = Mock()
        second = Mock()
        X = Mock()
        y = Mock()

        stacker = FeatureStacker([
            ('first', first),
            ('second', second),
            ])

        stacker.fit(X, y)
        first.fit.assert_called_with(X, y)
        second.fit.assert_called_with(X, y)

    def test_transform(self):
        from ..model import FeatureStacker

        first = Mock()
        second = Mock()
        X = Mock()

        stacker = FeatureStacker([
            ('first', first),
            ('second', second),
            ])

        value = stacker.transform(X)
        first.transform.assert_called_with(X)
        second.transform.assert_called_with(X)
        assert (value == [
            first.transform.return_value,
            second.transform.return_value,
            ]).all()

    def test_transform_with_sparse(self):
        from ..model import FeatureStacker

        first = Mock()
        second = Mock()
        X = Mock()

        first.transform.return_value = sparse.csr_matrix([
            [1, 2, 3], [3, 2, 1]])
        second.transform.return_value = sparse.csr_matrix([
            [4, 5, 6], [6, 5, 4]])

        stacker = FeatureStacker([
            ('first', first),
            ('second', second),
            ])

        value = stacker.transform(X)
        first.transform.assert_called_with(X)
        second.transform.assert_called_with(X)
        assert (value.toarray() == np.array([
            [1, 2, 3, 4, 5, 6],
            [3, 2, 1, 6, 5, 4],
            ])).all()

    def test_transform_with_sparse_single_feature(self):
        from ..model import FeatureStacker

        first = Mock()
        X = Mock()

        first.transform.return_value = sparse.csr_matrix([
            [1, 2, 3], [3, 2, 1]])

        stacker = FeatureStacker([
            ('first', first),
            ])

        value = stacker.transform(X)
        first.transform.assert_called_with(X)
        assert (value.toarray() == np.array([
            [1, 2, 3],
            [3, 2, 1],
            ])).all()

    def test_get_params(self):
        from ..model import FeatureStacker

        first = Mock()
        first.get_params.return_value = {
            'param1': 'value1', 'param2': 'value2'}

        stacker = FeatureStacker([
            ('first', first),
            ])

        value = stacker.get_params()
        assert value == {
            'first__param1': 'value1',
            'first__param2': 'value2',
            'first': first,
            }
        first.get_params.assert_called_with(deep=True)

    def test_get_params_deep_false(self):
        from ..model import FeatureStacker

        first = Mock()
        first.get_params.return_value = {
            'param1': 'value1', 'param2': 'value2'}

        stacker = FeatureStacker([
            ('first', first),
            ])

        value = stacker.get_params(deep=False)
        assert value == {
            'estimators': [('first', first)],  # ?
            }
        first.get_params.call_count == 0



import os

from setuptools import setup, find_packages

version = '0.1b1'

here = os.path.abspath(os.path.dirname(__file__))
try:
    README = open(os.path.join(here, 'README.rst')).read()
except IOError:
    README = ''

install_requires = [
    'docopt',
    'scikit-learn',
    ]

tests_require = [
    'mock',
    'pytest',
    'pytest-cov',
    ]

docs_require = [
    'Sphinx',
    ]

setup(name='nolearn',
      version=version,
      description="Miscellaneous utilities for machine learning.",
      long_description=README,
      classifiers=[
          'Development Status :: 4 - Beta',
        ],
      keywords='',
      author='Daniel Nouri',
      author_email='daniel.nouri@gmail.com',
      url='https://github.com/dnouri/nolearn',
      license='MIT',
      packages=find_packages(),
      include_package_data=True,
      zip_safe=False,
      install_requires=install_requires,
      extras_require={
          'testing': tests_require,
          'docs': docs_require,
          },
      )



import numpy as np
from sklearn.base import clone
from sklearn.metrics import f1_score


def multiclass_logloss(actual, predicted, eps=1e-15):
    """Multi class version of Logarithmic Loss metric.

    :param actual: Array containing the actual target classes
    :param predicted: Matrix with class predictions, one probability per class
    """
    # Convert 'actual' to a binary array if it's not already:
    if len(actual.shape) == 1:
        actual2 = np.zeros((actual.shape[0], predicted.shape[1]))
        for i, val in enumerate(actual):
            actual2[i, val] = 1
        actual = actual2

    clip = np.clip(predicted, eps, 1 - eps)
    rows = actual.shape[0]
    vsota = np.sum(actual * np.log(clip))
    return -1.0 / rows * vsota


class LearningCurve(object):
    score_func = staticmethod(f1_score)

    def __init__(self, score_func=None):
        if score_func is None:
            score_func = self.score_func
        self.score_func = score_func

    def predict(self, clf, X):
        return clf.predict(X)

    def __call__(self, dataset, classifier, steps=10,
                 verbose=0, random_state=42):
        """Create a learning curve that uses more training cases with
        each step.

        :param dataset: Dataset to work with
        :type dataset: :class:`~nolearn.dataset.Dataset`
        :param classifier: Classifier for fitting and making predictions.
        :type classifier: :class:`~sklearn.base.BaseEstimator`
        :param steps: Number of steps in the learning curve.
        :type steps: int

        :result: 3-tuple with lists `scores_train`, `scores_test`, `sizes`

        Drawing the resulting learning curve can be done like this:

        .. code-block:: python

          dataset = Dataset()
          clf = LogisticRegression()
          scores_train, scores_test, sizes = learning_curve(dataset, clf)
          pl.plot(sizes, scores_train, 'b', label='training set')
          pl.plot(sizes, scores_test, 'r', label='test set')
          pl.legend(loc='lower right')
          pl.show()
        """
        X_train, X_test, y_train, y_test = dataset.train_test_split()

        scores_train = []
        scores_test = []
        sizes = []

        if verbose:
            print "          n      train      test"

        for frac in np.linspace(0.1, 1.0, num=steps):
            frac_size = X_train.shape[0] * frac
            sizes.append(frac_size)
            X_train1 = X_train[:frac_size]
            y_train1 = y_train[:frac_size]

            clf = clone(classifier)
            clf.fit(X_train1, y_train1)

            predict_train = self.predict(clf, X_train1)
            predict_test = self.predict(clf, X_test)

            score_train = self.score_func(y_train1, predict_train)
            score_test = self.score_func(y_test, predict_test)

            scores_train.append(score_train)
            scores_test.append(score_test)

            if verbose:
                print "   %8d     %0.4f    %0.4f" % (
                    frac_size, score_train, score_test)

        return scores_train, scores_test, sizes


class LearningCurveProbas(LearningCurve):
    score_func = staticmethod(multiclass_logloss)

    def predict(self, clf, X):
        return clf.predict_proba(X)

learning_curve = LearningCurve().__call__
#: Same as :func:`learning_curve` but uses :func:`multiclass_logloss`
#: as the loss funtion.
learning_curve_logloss = LearningCurveProbas().__call__



from sklearn.linear_model import LogisticRegression
from sklearn.grid_search import GridSearchCV
from mock import Mock
import numpy as np


def test_grid_search():
    from ..grid_search import grid_search

    dataset = Mock()
    dataset.data = np.array([[1, 2, 3], [3, 3, 3]] * 20)
    dataset.target = np.array([0, 1] * 20)
    pipeline = LogisticRegression()
    parameters = dict(C=[1.0, 3.0])

    result = grid_search(dataset, pipeline, parameters)
    assert isinstance(result, GridSearchCV)
    assert hasattr(result, 'best_estimator_')
    assert hasattr(result, 'best_score_')



SAMPLE_SCHEMA = """
[first]
value1 = int
value2 = string
value3 = float
value4 = listofstrings
value5 = listofints

[second]
value1 = string
value2 = int
"""

SAMPLE_CONFIGURATION = """
[first]
value1 = 3
value2 = Three
value3 = 3.0
value4 = Three Drei
value5 = 3 3

[second]
value1 =
    a few line breaks

    are no problem

    neither is a missing value2
"""


def test_parse_config():
    from ..console import parse_config
    result = parse_config(SAMPLE_SCHEMA, SAMPLE_CONFIGURATION)
    assert result['first']['value1'] == 3
    assert result['first']['value2'] == u'Three'
    assert result['first']['value3'] == 3.0
    assert result['first']['value4'] == [u'Three', u'Drei']
    assert result['first']['value5'] == [3, 3]
    assert result['second']['value1'] == (
        u'a few line breaks\nare no problem\nneither is a missing value2')
    assert 'value2' not in result['second']
